package agents;
/*
 * (c)2010 Paul Smaldino at the University of California, Davis
 * This is an extension of EG_PublicGoods2 with Reputation. In this version, agents keep a list
 * of their co-players and remember past actions. This influences who they will let join games. 
 * 
 * This version introduces computations to calculate the assortment of the agents,
 * Above the expected level of assortment. 
 * 
 * 4/24/11
 * This new extension will implement a slightly more sophisticated version of reputation. 
 * The image scores will be held in a central matrix, this should conserve computational speed somewhat. 
 * 
 * This model is described in two papers: 
 * 1. Smaldino, PE, & Lubell, M. (2011). An institutional mechanism for assortment in an ecology of games. 
 * 		PLoS One, 6, e23019. 
 * 		This paper introduces capacity constraints.
 * 
 * 2. Smaldino, PE, & Lubell, M. (in review). Institutions and cooperation in an ecology of games. 
 * 		Artificial Life. 
 * 		This paper introduced reputational exclusion and compares the two mechanisms. 
 * 
 */

import agents.states.SimStateWithSimController;
import agents.control.*;
//import sim.field.continuous.Continuous2D;
import sim.field.grid.IntGrid2D;
import sim.field.grid.SparseGrid2D;
import sim.field.network.Network;
import sim.portrayal.Portrayal;
import sim.portrayal.continuous.ContinuousPortrayal2D;
import sim.portrayal.grid.SparseGridPortrayal2D;
import sim.util.Bag;
import sim.util.Double2D;
import sim.util.Int2D;
import sim.util.IntBag;


public class AgentsSimulation extends SimStateWithSimController { //extend the class SimState
	
	/********* Basic properties of the space and how agents interact with it ***********/
	//public SparseGrid2D agentsSpace;
	public SparseGridPortrayal2D agentsPortrayal = new SparseGridPortrayal2D();
	public int gridWidth = 30;
	public int gridHeight = 30;
	
	public IntGrid2D ecologyGrid;
	
	public SparseGrid2D agentsSpace; //holds the agents
	public SparseGrid2D gamesSpace; //holds the games
    public Network links; //the network, this will hold all the edges.
    GameEcology ge;
    
    public boolean[][] gameMatrix; //who is in which games
    public double gameProduction = 2.5;
    public boolean equalGameProductions = true;
    public boolean[][] coPlayerMatrix; //a matrix of who is a co-player with who
    
    public boolean stopAtEquilibrium = false;
    
    /********* Reputation Mechanism ***********/
    public double[][] reputation; //reputation[i][j] will be what i thinks of j, +1 for each cooperation, -1 for each defection.
	//public double repDiscount = 1.0; //lower values (0,1] mean that reputation builds more slowly.
	public boolean useReputation = false;
	public boolean outOfSightOutOfMind = false;//if true, agents start to forget less recent interactions.
	public double reputationThreshold = 0.5; //what fraction of players must vote YES to allow an agent to join a game?
	public double repConvergenceSlope = 1.0;// larger values means faster convergence.
	public double repErrorSD = 1.0; // the SD of uniformly distributed error in the reputation. 
	public boolean multipleObservances = false; //does each observation in each game count?
	public double biasShift = 0.0; //How much do cooperators assume more cooperative, defectors less so
	
	/********* Agent Properties ***********/
	public int numAgents = 100; //how many agents? 
	public double initialEndowment = 10.0;
	public int numGames = 100; //how many games?
	public double probJoinGameInit = 0.1; // probability of joining each game initially
	public double probJoin = 0.10; //probability of joining a new game
	public double probLeave = 0.10; //probability of leaving an old game
	public double coopFreq = 0.5; //the initial frequency of cooperators.
	public int maxGames = 100; //the maximum number of games an agent can be playing at once. 
	public int maxPlayers = 100; //the maximum number of players a game can host.
	
	/********* Agent Strategies ***********/
	public final static int DEFECTOR = 0;
	public final static int COOPERATOR = 1;
	public double contribution = 1.0;//how much a cooperator contributes.
	public boolean distributedContribution = true; //C contributes c/k if true, c otherwise. 
	
	public double[] coopPayoffs; //the payoffs to all the cooperating agents.
	public double[] defectPayoffs; //the payoffs to all the defecting agents.
	
	
	/********* Switching Rules ***********/
	public final static int mostPlayers = 0;
	public final static int leastPlayers = 0;
	public final static int highestPayoff = 1;
	public final static int lowestPayoff = 1;
	public int leaveRule = lowestPayoff;
	public int joinRule = mostPlayers;
	public double joinRandomGame = 0.0; // probability of choosing a random game to join.
	public double leaveRandomGame = 0.0;//probability of choosing a random game to leave.
	public boolean joinTheBest = false;//when joining, strictly pick the one with highest payoff
	public boolean leaveOnlyLosingGames = true; //only leave games with payoff less than contribution
	
	public boolean joinGamesOnlyInNetwork = true;
	
	public boolean histograms = false; //make the histograms?
	public boolean fancyColors = true; //make fancy colors for display?
	
	/********* Variables for Printing Data ***********/
	//public boolean printScreen = true;
	public int printSteps = 20;
	public boolean printToConsole = true;
	public boolean printDataRows = true;//prints the headers on the file for each run.
	public String dataFileName = "data.txt";
	public String dataFolderName = "data";
	public boolean saveDataToFile = false; //save the data to file(s)?
	public boolean savePNGToFile = false; //snap pictures automatically? 
	public String simulationFolder	= "";  			//folder that contains all of the files for a simulation
	public String parametersFile = "";   //file with simulation parameters
	public String dataFilePrefix;				// prepended to the number of the run
	public String dataFileFileType;					// file type saved	
	public String picFilePrefix = "picture";	// prepended to the number of the run
	public String picFileFileType = "png";					// file type saved
	
	public long lengthOfSimulations = 1000000; //how many steps in a run? (maximum)
	public int numberOfSimulations; 	// The number of Simulations
	public int simulationCount = 0;
	
	public boolean printNetworkData = false;//output network info files?
	public boolean printGameInfo = false; //output gameData (C and D players at beginning and end) to a file?
	public double[][] gameInfo;
	
	public double timePastRCP;
	
	
	
	
	/********* Crap that I need ***********/
	Observer ob = null; //observer agent
	public SimController sc = null;
	public boolean doGraphics = true; //set to false for batch runs, saves time. 
	public Bag agentBag = new Bag();
	public boolean autoQuit = false;
	public Bag recycledAgents = new Bag();
	public int simCount = 0; //what's the number of the current run?
	

	/*
	 * *******************Get and Set Methods*****************************************
	 */
	public boolean getautoSim(){return autoSimulation;}
	public void setautoSim(boolean b){autoSimulation = b;}
	
   /* public double getRCP(){
    	double payC = average(this.coopPayoffs);
    	double payD = average(this.defectPayoffs);
    	return payC/payD;
    }*/
	
	public int getgridWidth (){   	return gridWidth;        }
    public void setgridWidth(int w){    if(w > 0){	gridWidth = w; gridHeight = w;     }  }
    public int getnumAgents (){   	return numAgents;        }
    public void setnumAgents(int w){    if(w > 0)	numAgents = w;      }
    public int getmaxPlayers (){   	return maxPlayers;        }
    public void setmaxPlayers(int w){    if(w > 0 && w <= numAgents)	maxPlayers = w;      }
    public int getnumGames (){   	return numGames;        }
    public void setnumGames(int w){    if(w > 0)	numGames = w;      }
    public int getmaxGames (){   	return maxGames;        }
    public void setmaxGames(int w){    if(w > 0 && w <= numGames)	maxGames = w;      }
    public double getcoopFreq (){   	return coopFreq;        }
    public void setcoopFreq(double w){    if(w >= 0 && w <= 1)	coopFreq = w;      }
    public double getprobJoinGameInit (){   	return probJoinGameInit;        }
    public void setprobJoinGameInit(double w){    if(w >= 0 && w <= 1)	probJoinGameInit = w;      }
    public double getprobJoin (){   	return probJoin;        }
    public void setprobJoin(double w){    if(w >= 0 && w <= 1)	probJoin = w;      }
    public double getprobLeave (){   	return probLeave;        }
    public void setprobLeave(double w){    if(w >= 0 && w <= 1)	probLeave = w;      }
    
    public double getjoinRandomGame (){   	return joinRandomGame;        }
    public void setjoinRandomGame(double w){    if(w >= 0 && w <= 1)	joinRandomGame = w;      }
    public double getleaveRandomGame (){   	return leaveRandomGame;        }
    public void setleaveRandomGame(double w){    if(w >= 0 && w <= 1)	leaveRandomGame = w;      }
  //  public boolean getjoinTheBest (){   	return joinTheBest;        }
  //  public void setjoinTheBest(boolean w){    	joinTheBest = w;      }
    public boolean getjoinGamesOnlyInNetwork (){   	return joinGamesOnlyInNetwork;        }
    public void setjoinGamesOnlyInNetwork(boolean w){    	joinGamesOnlyInNetwork = w;      }
    public boolean getleaveOnlyLosingGames (){   	return leaveOnlyLosingGames;        }
    public void setleaveOnlyLosingGames(boolean w){    	leaveOnlyLosingGames = w;      }
    
    public boolean getdistributedContribution (){   	return distributedContribution;        }
    public void setdistributedContribution(boolean w){    	distributedContribution = w;      }
    public double getgameProduction (){   	return gameProduction;        }
    public void setgameProduction(double w){    if(w >= 1)	gameProduction = w;      }
    public boolean getequalGameProductions (){   	return equalGameProductions;        }
    public void setequalGameProductions(boolean w){    	equalGameProductions = w;      }
    //*****************REPUTATION******************************
    public boolean getuseReputation (){   	return useReputation;        }
    public void setuseReputation(boolean w){    	useReputation = w;      }
    public double getreputationThreshold (){   	return reputationThreshold;        }
    public void setreputationThreshold(double w){    if(w >= 0 && w <= 1.0)	reputationThreshold = w;      }
    public double getrepConvergenceSlope (){   	return repConvergenceSlope;        }
    public void setrepConvergenceSlope(double w){    if(w > 0)	repConvergenceSlope = w;      }
    public double getrepErrorSD (){   	return repErrorSD;        }
    public void setrepErrorSD(double w){    if(w >= 0)	repErrorSD = w;      }
    public boolean getmultipleObservances (){   	return multipleObservances;        }
    public void setmultipleObservances(boolean w){    	multipleObservances = w;      }
    public boolean getoutOfSightOutOfMind (){   	return outOfSightOutOfMind;        }
    public void setoutOfSightOutOfMind(boolean w){    	outOfSightOutOfMind = w;      }
    public double getbiasShift(){   	return biasShift;        }
    public void setbiasShift(double w){   	biasShift = w;      }
    //*****************FILE I/O******************************
	public String getdataFileName(){		return dataFileName; }
	public void setdataFileName(String s){		if(s.length() > 0)	dataFileName = s; }
	public String getdataFolderName(){		return dataFolderName; }
	public void setdataFolderName(String s){		if(s.length() > 0)		dataFolderName = s; }
	public boolean getsaveDataToFile(){		return saveDataToFile;	}
	public void setsaveDataToFile(boolean b){		saveDataToFile = b;	}
	public boolean getprintToConsole(){            return printToConsole;        }        
    public void setprintToConsole(boolean b){    printToConsole = b;        }
	public int getprintSteps(){ return printSteps;}
	public void setprintSteps(int h){ if (h > 0) printSteps = h;}
	public boolean getdoGraphics(){            return doGraphics;        }        
    public void setdoGraphics(boolean b){    doGraphics = b;        }
    public boolean gethistograms(){            return histograms;        }        
    public void sethistograms(boolean b){    histograms = b;        }
    public boolean getfancyColors(){            return fancyColors;        }        
    public void setfancyColors(boolean b){    fancyColors = b;        }
    public boolean getstopAtEquilibrium (){   	return stopAtEquilibrium;        }
    public void setstopAtEquilibrium(boolean w){    	stopAtEquilibrium = w;      }

	
    /**
	 * returns the mean of an array of doubles
	 */
	public double average(double[] vector){
		double sum = 0;
		final double size = (double)vector.length;
		for(int i = 0; i < size; i++)
			sum += vector[i];
		return sum/size;
	}
	
	
	/*********************	Methods for Histograms	********************************************/
	public double[] gameSizes() {//makes histogram with how many players per game
        try
            {
        	double[] d = new double[numGames];
            Bag bag = gamesSpace.getAllObjects();
            for(int i = 0; i < numGames; i++){
            	Game g = (Game)bag.get(i);
            	d[i] = Math.round(g.howManyPlayers(this));
            }
            return d;
            }	
        catch (Exception e) { return new double[1]; }	 
        }	
	
	public double[] gameSizesExclusive() {//Same as above, but excludes empty games
        try
            {
        	int ng = numGames;
        	int count;
        	for(int i = 0; i < numGames; i++){ //ignore empty games
        		count = 1;
        		for(int j = 0; j < numAgents; j++){
        			if(gameMatrix[j][i]){
        				count = 0;
        				break;
        			}
        		}
        		ng -= count;
        	}
    		double[] d = new double[ng];
            Bag bag = gamesSpace.getAllObjects();
            int index = 0;
            for(int i = 0; i < numGames; i++){
            	Game g = (Game)bag.get(i);
            	int x = g.howManyPlayers(this);
            	if(x > 0){
            		d[index] = Math.round(x);
            		index++;
            	}
            }
            return d;
            }	
        catch (Exception e) { return new double[1]; }	 
        }	
	
	public double[] coopGameSizesExclusive() {//how many cooperators in each game.
		try
		{
	    	int ng = numGames;
	    	int count;
	    	for(int i = 0; i < numGames; i++){ //ignore empty games
	    		count = 1;
	    		for(int j = 0; j < numAgents; j++){
	    			if(gameMatrix[j][i]){
	    				count = 0;
	    				break;
	    			}
	    		}
	    		ng -= count;
	    	}
	    	
			double[] d = new double[ng];
	        Bag games = gamesSpace.getAllObjects();
	        Bag agents = agentsSpace.getAllObjects();
	        int index = 0;
	        for(int i = 0; i < numGames; i++){
	        	Game g = (Game)games.get(i);
	        	int n = g.howManyPlayers(this);
	        	int c = 0;
	        	if(n > 0){	
	        		for(int j = 0; j < numAgents; j++){
	        			if(gameMatrix[j][i]){
	        				Agent a = (Agent)agents.get(j);
	        				if(a.strategy == COOPERATOR)
	        					c++;
	        			}
	        		}
	        		d[index] = (double)c;
	        		index++;
	        	}
	        }
	        return d;
	        }	
	    catch (Exception e) { return new double[1]; }	
        }	
	
	
	public double[] defectGameSizesExclusive() {//how many defectors in each game.
		try
		{
	    	int ng = numGames;
	    	int count;
	    	for(int i = 0; i < numGames; i++){ //ignore empty games
	    		count = 1;
	    		for(int j = 0; j < numAgents; j++){
	    			if(gameMatrix[j][i]){
	    				count = 0;
	    				break;
	    			}
	    		}
	    		ng -= count;
	    	}
	    	
			double[] d = new double[ng];
	        Bag games = gamesSpace.getAllObjects();
	        Bag agents = agentsSpace.getAllObjects();
	        int index = 0;
	        for(int i = 0; i < numGames; i++){
	        	Game g = (Game)games.get(i);
	        	int n = g.howManyPlayers(this);
	        	int c = 0;
	        	if(n > 0){	
	        		for(int j = 0; j < numAgents; j++){
	        			if(gameMatrix[j][i]){
	        				Agent a = (Agent)agents.get(j);
	        				if(a.strategy == DEFECTOR)
	        					c++;
	        			}
	        		}
	        		d[index] = (double)c;
	        		index++;
	        	}
	        }
	        return d;
	        }	
	    catch (Exception e) { return new double[1]; }	
        }	
	

public double[] agentConnectivity(){//makes histogram with how many games per player
	try
	{
    	double[] d = new double[numAgents];
        Bag bag = agentsSpace.getAllObjects();
        for(int i = 0; i < numAgents; i++){
        	Agent a = (Agent)bag.get(i);
        	d[i] = Math.round(a.howManyGames(this));
        }
        return d;
        }	
    catch (Exception e) { return new double[1]; }	
}
public double[] coopConnectivity(){//makes histogram with how many games per cooperator
	try
	{
		int numCoop = (int)((double)numAgents * coopFreq);
    	double[] d = new double[numCoop];
        Bag bag = agentsSpace.getAllObjects();
        int index = 0;
        for(int i = 0; i < numAgents; i++){
        	Agent a = (Agent)bag.get(i);
        	if(a.strategy == COOPERATOR){
        		d[index] = Math.round(a.howManyGames(this));
        		index++;
        	}
        }
        return d;
        }	
    catch (Exception e) { return new double[1]; }	
}

public double[] defectConnectivity(){//makes histogram with how many games per defector
	try
	{
		int numDefect = numAgents - (int)((double)numAgents * coopFreq);
    	double[] d = new double[numDefect];
        Bag bag = agentsSpace.getAllObjects();
        int index = 0;
        for(int i = 0; i < numAgents; i++){
        	Agent a = (Agent)bag.get(i);
        	if(a.strategy == DEFECTOR){
        		d[index] = Math.round(a.howManyGames(this));
        		index++;
        	}
        }
        return d;
        }		
    catch (Exception e) { return new double[0]; }	
}

public int activeGames(){//returns the number of active games.
	try{
		int x = 0;
		for(int i = 0; i < numGames; i++){
			for(int j = 0; j < numAgents; j++)
				if(gameMatrix[j][i]){
					x++;
					break;
				}
		}
		return x;
	}
	catch (Exception e) { return 0;}	
}

/**
 * for a histogram of cooperator Frequencies in each game.
 */
public double[] gameCoopFreqs(){ 
	try
	{
    	int ng = numGames;
    	int count;
    	for(int i = 0; i < numGames; i++){ //ignore empty games
    		count = 1;
    		for(int j = 0; j < numAgents; j++){
    			if(gameMatrix[j][i]){
    				count = 0;
    				break;
    			}
    		}
    		ng -= count;
    	}
    	
		double[] d = new double[ng];
        Bag games = gamesSpace.getAllObjects();
        Bag agents = agentsSpace.getAllObjects();
        int index = 0;
        for(int i = 0; i < numGames; i++){
        	Game g = (Game)games.get(i);
        	int n = g.howManyPlayers(this);
        	int c = 0;
        	if(n > 0){	
        		for(int j = 0; j < numAgents; j++){
        			if(gameMatrix[j][i]){
        				Agent a = (Agent)agents.get(j);
        				if(a.strategy == COOPERATOR)
        					c++;
        			}
        		}
        		d[index] = (double)c/(double)n;
        		index++;
        	}
        }
        return d;
        }	
    catch (Exception e) { return new double[1]; }	
}


/**
 * for a histogram of defector Frequencies in each game.
 */
public double[] gameDefectorFreqs(){ 
	try
	{
    	int ng = numGames;
    	int count;
    	for(int i = 0; i < numGames; i++){ //ignore empty games
    		count = 1;
    		for(int j = 0; j < numAgents; j++){
    			if(gameMatrix[j][i]){
    				count = 0;
    				break;
    			}
    		}
    		ng -= count;
    	}
    	
		double[] d = new double[ng];
        Bag games = gamesSpace.getAllObjects();
        Bag agents = agentsSpace.getAllObjects();
        int index = 0;
        for(int i = 0; i < numGames; i++){
        	Game g = (Game)games.get(i);
        	int n = g.howManyPlayers(this);
        	int c = 0;
        	if(n > 0){	
        		for(int j = 0; j < numAgents; j++){
        			if(gameMatrix[j][i]){
        				Agent a = (Agent)agents.get(j);
        				if(a.strategy == DEFECTOR)
        					c++;
        			}
        		}
        		d[index] = (double)c/(double)n;
        		index++;
        	}
        }
        return d;
        }	
    catch (Exception e) { return new double[1]; }	
}
	
/* ************************** end Methods for Histograms*********************************************/ 

public double[] gamePayouts(){
	try
	{
    	int ng = numGames;
    	int count;
    	for(int i = 0; i < numGames; i++){ //ignore empty games
    		count = 1;
    		for(int j = 0; j < numAgents; j++){
    			if(gameMatrix[j][i]){
    				count = 0;
    				break;
    			}
    		}
    		ng -= count;
    	}
    	
		double[] d = new double[ng];
        Bag games = gamesSpace.getAllObjects();
        //Bag agents = agentsSpace.getAllObjects();
        int index = 0;
        for(int i = 0; i < numGames; i++){
        	Game g = (Game)games.get(i);
        	int n = g.howManyPlayers(this);
        	int c = 0;
        	if(n > 0){	
        		
        		d[index] = g.payout;
        		index++;
        	}
        }
        return d;
        }		
    catch (Exception e) { return new double[1]; }	
	
}






	
	/**
	 * Create the grids and spaces
	 */
	public void createGrids(){
		agentsSpace = new SparseGrid2D(gridWidth,gridHeight);
		gamesSpace = new SparseGrid2D(gridWidth,gridHeight);
		links = new Network();
		
		ecologyGrid = new IntGrid2D(numAgents,numGames);
		
	}

	public void initEcoGrid(){
		Bag agents = agentsSpace.getAllObjects();
		int strategy = 0;
		for(int x = 0; x < numAgents; x++){
			Agent a = (Agent)agents.get(x);
			strategy = (-1*a.strategy) + 2; //1 for coop, 2 for defector.
			IntBag games = a.currentGamesInd(this);	
			for(int j = 0; j < games.numObjs; j++){
				int y = games.get(j);
				ecologyGrid.set(x, y, strategy);
			}	
		}
	}
	
	
	/**
	 * Creates the agents and places them on the grid.
	 */
	public void createAgents(){
		int numCoop = (int)((double)numAgents * coopFreq);
		int numDefect = numAgents - numCoop;
		coopPayoffs = new double[numCoop];
		defectPayoffs = new double[numDefect];
	   for(int i = 0; i < numCoop; i++){ //create cooperators
		   final Agent agent = new Agent(initialEndowment, numGames);
		   agent.strategy = COOPERATOR;
		   agentsSpace.setObjectLocation(agent, new Int2D(1 + (random.nextInt(gridWidth-2)),
	                  1 +  (random.nextInt(gridHeight-1))));
		   links.addNode(agent);
		   agent.event = schedule.scheduleRepeating(agent, 2, 1);
	    }
	   for(int i = 0; i < numDefect; i++){ //create defectors
		   final Agent agent = new Agent(initialEndowment, numGames);
		   agent.strategy = DEFECTOR;
		   agentsSpace.setObjectLocation(agent, new Int2D(1 + (random.nextInt(gridWidth-2)),
	                  1 +  (random.nextInt(gridHeight-1))));
		   links.addNode(agent);
		   agent.event = schedule.scheduleRepeating(agent, 2, 1);
	    }
	}
	
	/**
	 * Creates the games and places them on the grid.
	 */
	public void createGames(){
		for(int i=0; i < numGames; i++){
			double p = (equalGameProductions) ? gameProduction: random.nextDouble()*(gameProduction-1) + 1; //production function is random between 1-2.
			final Game game = new Game(p, this);
			gamesSpace.setObjectLocation(game, new Int2D(1 + (random.nextInt(gridWidth-2)),
	                  1 +  (random.nextInt(gridHeight-1))));
			links.addNode(game);
		}
	}
	
	
	/**
	 * Let each agent know its index number, and let it be identical to its index in 
	 * agentsSpace. 
	 */
	public void indexAgents(){
		Bag agents = agentsSpace.getAllObjects();
		for(int i = 0; i < agents.numObjs; i++){
			Agent a = (Agent)agents.get(i);
			a.index = i;
		}
	}
	
	
	
	
	/**
	 * Initial assignment of agents to participate in the games
	 * PROBLEM: When maxPlayers < numAgents, there is a disparity in the number of games
	 * initially assigned to cooperators and defectors. 
	 * This has been fixed: use assignGames2(). 
	 */
	public void assignGames(){
		Bag agents = agentsSpace.getAllObjects();
		Bag games = gamesSpace.getAllObjects();
		boolean gameless = true;
		for(int i = 0; i < numAgents; i++){
			Agent a = (Agent)agents.get(i);
			gameless = true;
			for(int j = 0; j < numGames; j++){
				Game g = (Game)games.get(j);
				if(random.nextBoolean(probJoinGameInit) && a.howManyGames(this) < maxGames && 
						g.howManyPlayers(this) < maxPlayers){
					links.addEdge(a, g, 1); //add a link from a to g
					gameless = false;
					gameMatrix[a.index][g.index] = true;
				}
			}
			if(gameless){//if agent hasn't joined any games, have it join a game
				Game g = (Game)games.get(random.nextInt(numGames));
				//while(g.howManyPlayers(this) >= maxPlayers)//a good idea, but can lead to endless loops.
				//	g = (Game)games.get(random.nextInt(numGames));//it's ok - the agents sometimes start with more per game, but tend to even out.
				links.addEdge(a, g, 1);
				gameMatrix[a.index][g.index] = true;
			}	
		}
	}
	
	/**
	 * A better method for assigning games. 
	 * The idea is to create a list of all the agent-game pairings, 
	 * and then randomize the list. We then go through the list to give each pairing a basically
	 * equal shot at happening, and there should be no bias for the combos at the end of the list
	 * versus the beginning, so we don't get effects of order for number of games joined, etc.  
	 */
	public void assignGames2(){
		Bag agents = agentsSpace.getAllObjects();
		Bag games = gamesSpace.getAllObjects();
//Create a vector of every agent combination.
		Int2D[] combos = new Int2D[numGames * numAgents];
		int[] indices = new int[numGames * numAgents];
		int x = 0;
		for(int i = 0; i < numAgents; i++)
			for(int j = 0; j < numGames; j++){
				combos[x] = new Int2D(i, j);
				indices[x] = x;
				x++;
			}
//Create a vector of the indices in random order
		indices = shuffle(indices);
//Probabilistically assign agents to games.
		for(int i = 0; i < indices.length; i++){
			if(random.nextBoolean(probJoinGameInit)){
				Agent a = (Agent)agents.get(combos[indices[i]].x);
				Game g = (Game)games.get(combos[indices[i]].y);
				if(a.howManyGames(this) < maxGames && g.howManyPlayers(this) < maxPlayers){
					links.addEdge(a, g, 1); //add a link from a to g
					gameMatrix[a.index][g.index] = true;
				}
			}
		}
//Make sure no agents are gameless.
		boolean search = (maxPlayers*numGames > numAgents) ? true: false;
		for(int i = 0; i < numAgents; i++){
			if(gameless(i)){
				Agent a = (Agent)agents.get(i);
				Game g = (Game)games.get(random.nextInt(numGames));
				int count = 1;
				while(search && g.howManyPlayers(this) >= maxPlayers && count <= numGames){
					g = (Game)games.get(random.nextInt(numGames));
					//System.out.println("finding a new game for agent " + a.index + "attempt #" + count);
					count++;
				}
				links.addEdge(a, g, 1);
				gameMatrix[a.index][g.index] = true;
			}
		}
		if(maxPlayers < numAgents)
			winnowGames();
		
	}
	
	
	/**
	 * Yet another way to assign agents to games. This method keeps the initial number of games per agent as even as possible,
	 * which may be a bad idea, but eliminates the problem with too few games and limitations on max agents-per-game,
	 * where most agents are only in one game (and therefore can't leave). 
	 */
	public void assignGames3(){
		int maxGame = (int)Math.floor(probJoinGameInit*numGames); //how many games should each agent join? 
		if(maxGame > this.maxGames)
			maxGame = this.maxGames;
		//System.out.println("maxGame = " + maxGame);
		Bag agents = agentsSpace.getAllObjects();
		Bag games = gamesSpace.getAllObjects();
		int[] indices = new int[numAgents];
		for(int i = 0; i < numAgents; i++)
			indices[i] = i;
		indices = shuffle(indices);
		Game gf = (Game)games.get(numGames - 1); //last game
		Agent af = (Agent)agents.get(indices[numAgents -1]); //last agent
		int gameIndex = 0;
		int count = 0;
		while(af.howManyGames(this) < maxGame && gf.howManyPlayers(this) < maxPlayers && count < 100){
			gameIndex = random.nextInt(numGames - 1);//start at a random game for each pass.
			for(int i = 0; i < numAgents; i++){ //for each agent
				int x = indices[i]; //in random order.
				Agent a = (Agent)agents.get(x);
				Game g = (Game)games.get(gameIndex);
				
				if(a.howManyGames(this) < maxGame && g.howManyPlayers(this) < maxPlayers){
					gameMatrix[x][gameIndex] = true;
					links.addEdge(a, g, 1);
					gameIndex++;
					if(gameIndex >= numGames)
						gameIndex = 0;
				}
				else{
					count++;
					break;
				}
			}	
		}
		if(maxPlayers < numAgents)
			winnowGames();
	}
	
	
	
	
	/**
	 * If there is a limit of max Agents per Game, we want to make sure there are not games that go over the limit. 
	 */
	public void winnowGames(){
		Bag games = gamesSpace.getAllObjects();
		Bag agents = agentsSpace.getAllObjects();
		for(int i = 0; i < numGames; i++){//go through all the games
			Game g = (Game)games.get(i);
			int count = 0;
			while(g.howManyPlayers(this) > maxPlayers && count < 1000){
				IntBag currPlayers = g.currentPlayers(this); //indices of the game's current players. 
				int x = random.nextInt(g.howManyPlayers(this));
				Agent a = (Agent)agents.get(currPlayers.get(x)); //pick a random player.
				if(a.howManyGames(this) > 1)
					gameMatrix[a.index][g.index] = false; //remove the agent from the game.	
				count++;
			}
		}	
	}
	
	
	
	/**
	 * Takes a vector and randomizes the order of the elements
	 */
	public int[] shuffle(int[] vector){
		int[] x = vector;
		final int n = x.length;
		int x1 = 0, x2 = 0, r = 0;
		for(int i = 0; i < n; i++){
			r = i + random.nextInt(n - i); //a random number between i and n-1.
			x1 = x[i];
			x2 = x[r];
			x[i] = x2;
			x[r] = x1;
		}
		return x;
	}
	
	/*
	 * Returns true if the agent with the given index is not playing ANY games.
	 */
	public boolean gameless(int index){
		boolean x = true;
		for(int j = 0; j < numGames; j++){
			if(gameMatrix[index][j]){
				x = false;
				break;
			}
		}
		return x;
	}
	
	
	/**
	 * Let each game and each agent know its own index
	 * The index in the Bag and the index in gamesSpace is the same.
	 */
	public void indexGames(){
		Bag games = gamesSpace.getAllObjects();
		for (int i = 0; i < numGames; i++){
			Game g = (Game)games.get(i);
			g.index = gamesSpace.getObjectIndex(g);
		}
		
		Bag agents = agentsSpace.getAllObjects(); //also do agents' index
		for(int i = 0; i < numAgents; i++){
			Agent a = (Agent)agents.get(i);
			a.index = agentsSpace.getObjectIndex(a);
		}
		
	}
	
	
	/**
	 * 
	 */
	public void gameInfoInit(){
		gameInfo = new double[numGames][8];//C and D pops, totPop, cFreq at beginning and end.
		Bag games = gamesSpace.getAllObjects();
		for(int i = 0; i < numGames; i++){
			Game g = (Game)games.get(i);
			int p = g.howManyPlayers(this);
			int c = g.howManyCooperators(this);
			gameInfo[i][0] = c;
			gameInfo[i][1] = p - c;
			gameInfo[i][2] = p;
			double cFreq = (p > 0) ? (double)c/(double)p : 0;
			gameInfo[i][3] = cFreq;
		}	
	}
	
	public void gameInfoLast(){
		Bag games = gamesSpace.getAllObjects();
		for(int i = 0; i < numGames; i++){
			Game g = (Game)games.get(i);
			int p = g.howManyPlayers(this);
			int c = g.howManyCooperators(this);
			gameInfo[i][4] = c;
			gameInfo[i][5] = p - c;
			gameInfo[i][6] = p;
			double cFreq = (p > 0) ? (double)c/(double)p : 0;
			gameInfo[i][7] = cFreq;
		}
		
	}
	

	/**
	 * Updates the co-player matrix, who is in a game with who.
	 * This is not the fastest way to do this, but that's OK.
	 */
	public void updateCoplayerMatrix(){
		for(int i = 0; i < numAgents; i++)
			for(int j = 0; j < numAgents; j++)
				for(int g = 0; g < numGames; g++){
					if(gameMatrix[i][g] && gameMatrix[j][g]){
						coPlayerMatrix[i][j] = true;
						break;
					}
					coPlayerMatrix[i][j] = false;
				}
	}
	
	
	/**
	 * creates an initial bias with trusting C's and skeptical D's. 
	 */
	public void initReputationMatrix(){
		Bag agents = agentsSpace.getAllObjects();
		for (int i = 0; i < numAgents; i++){
			Agent a = (Agent)agents.get(i);
			final int ind = a.index;
			if(a.strategy == AgentsSimulation.COOPERATOR){
				for(int j = 0; j < numAgents; j++)
					reputation[ind][j] += biasShift;
			}
			else{
				for(int j = 0; j < numAgents; j++)
					reputation[ind][j] -= biasShift;
			}
		}
	}
	
	
	
	/**
	 * Returns the logistic function of x
	 * repConvergenceSlope is the speed of convergence to 0 or 1, larger numbers converge faster
	 * repErrorSD is the SD of the normally distributed error.
	 */
	public double logistic(double x){
		final double error = random.nextGaussian()*repErrorSD;
		final double term = -1*repConvergenceSlope*(x + error);
		final double p = 1 + Math.exp(term);
		return (1/p);
	}
	
	/**
	 * Return the average assortment index. 
	 */
	public double calculateAssortment(){
		double assortment = 0;
		Bag agents = agentsSpace.getAllObjects();
		for(int i = 0; i < numAgents; i++){
			Agent a = (Agent)agents.get(i);
			assortment += a.assortmentIndex(this);
		}
		double obsAssrt = assortment/(double)numAgents; 
		
		final double f = coopFreq;
		double expAssrt = (f*f) + (1-f)*(1-f);
		return (obsAssrt - expAssrt);
	}
	
	
	/**
	 * 
	 */
	public double[][] gameInfoHist(){
		double[][] gamesCD = new double[2][numAgents + 1];
		Bag games = gamesSpace.getAllObjects();
		for(int i = 0; i < games.numObjs; i++){
			Game g = (Game)games.get(i);
			int c = g.howManyCooperators(this);
			int d = g.howManyPlayers(this) - c;
			gamesCD[0][c]++;
			gamesCD[1][d]++;
		}
		return gamesCD;
	}
	
	/**
	 * returns the relative cooperator payoff (RCP)
	 */
	public double returnRCP(){
		double payC = average(this.coopPayoffs);
		double payD = average(this.defectPayoffs);
		return payC/payD;
	}
	
	

	public AgentsSimulation(long seed) {
	    super(seed);
	    agentsSpace = new SparseGrid2D(gridWidth,gridHeight);
		gamesSpace = new SparseGrid2D(gridWidth,gridHeight);
		links = new Network(); 
	    }
	
	/*************** Start Method ********************************************/
	public void start(){
		super.start(); // reuse the SimState start method
		simCount++; //increment the simCount
		
			/** Stuff needed for batch runs		 **/
			checkReleaseAutomatedControl();
			sc = checkAutomatedControl(); // allow the observer and controller to interact
			checkReleaseAutomatedControl();
			recycledAgents.clear(); //clears the recycled agents from previous simulations. 
		
	    gameMatrix = new boolean[numAgents][numGames];
	    final int numCoop = (int)((double)numAgents * coopFreq);
		final int numDefect = numAgents - numCoop;
		coopPayoffs = new double[numCoop];
		defectPayoffs = new double[numDefect];
		coPlayerMatrix = new boolean[numAgents][numAgents];
	    
	    
		createGrids(); //set up the grid(s)
		createAgents(); //create the agents
		indexAgents();
		reputation = new double[numAgents][numAgents];
		if(biasShift != 0)
			initReputationMatrix();//create an initial bias with trusting C's and skeptical D's. 
		createGames(); //create the games
		indexGames();
		assignGames3(); //assign agents to games	//use assignGames2() for better randomizing.
		updateCoplayerMatrix();
		gameInfoInit();
		ge = new GameEcology();
		ge.event = schedule.scheduleRepeating(ge, 1, 1);
		
		timePastRCP = 0;
		
		if(doGraphics)
			initEcoGrid();
		

	    //Create the observer agent who will collect the data at the end of each time step.
	   ob = new Observer(this,ob, dataFileName , dataFolderName , saveDataToFile,savePNGToFile, sc,precision);

	}

	
	/**
	 * This method just needs to be here. Don't mess with it.
	 */
	public static void main(String[] args) {
		doLoop(AgentsSimulation.class, args); //doLoop is a static method 
		//already defined in SimState
		System.exit(0);//Stop the program when finished.
	}


}